/********************************************************************
 * thinkfan_system.c
 *
 * This work is licensed under a Creative Commons Attribution-Share Alike 3.0
 * United States License. See http://creativecommons.org/licenses/by-sa/3.0/us/
 * for details.
 *
 * This file contains all functions that are specific to dealing with
 * either /sys/class/hwmon or the /proc/acpi/ibm interface. They are
 * referenced in the main program via function pointers.
 *
 * This file is part of thinkfan. See thinkfan.c for further info.
 * ******************************************************************/

#define _GNU_SOURCE

#include <syslog.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include "thinkfan.h"

#define IBM_TEMP "/proc/acpi/ibm/thermal"
#define IBM_FAN "/proc/acpi/ibm/fan"
#define PID_FILE "/var/run/thinkfan.pid"

int get_temp_ibm();
int get_temp_sysfs();
void setfan_ibm(int);
void setfan_sysfs(int);
void setfan_sysfs_safe(int);
void init_fan_ibm();
void init_fan_sysfs();
void uninit_fan_ibm();
void uninit_fan_sysfs();

/*******************************************************************
 * get_temp_ibm reads temperatures from /proc/acpi/ibm/thermal and
 * returns the highest one found.
 *******************************************************************/
int get_temp_ibm() {
	int i=0, res, retval=0;
	int *t = malloc(16 * sizeof(int));
	int ibm_temp;
	ssize_t r;
	char buf[128];

	memset(t, 0, 16);
	if (((ibm_temp = open(IBM_TEMP, O_RDONLY)) < 0)
			|| ((r = read(ibm_temp, &buf, 128)) < 14)
			|| (close(ibm_temp) < 0)) {
		showerr(IBM_TEMP);
		free(t);
		errcnt++;
		return -1000;
	}
	buf[r] = 0;
	res = sscanf(buf,
	 "temperatures: %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i",
	 &t[0], &t[1], &t[2], &t[3], &t[4], &t[5], &t[6], &t[7], &t[8], &t[9],
	 &t[10], &t[11], &t[12], &t[13], &t[14], &t[15]);
	if (res < 2) {
		message(LOG_ERR, MSG_ERR_T_GET);
		free(t);
		errcnt++;
		return -1000;
	}
	for (i=0; i < 16; i++)
		if (t[i] > retval) retval = t[i];
	free(t);
	return retval;
}

/****************************************************************
 * get_temp_sysfs() reads the temperature from all files that
 * were specified as "sensor ..." in the config file and returns
 * the highest temperature.
 ****************************************************************/
int get_temp_sysfs() {
	int num, fd, idx = 0;
	long int rv = 0, tmp;
	char buf[7];
	char *endptr;
	while(config->sensors[idx] != NULL) {
		if ((fd = open(config->sensors[idx], O_RDONLY)) == -1) {
			showerr(config->sensors[idx]);
			errcnt++;
			return -1000;
		}
		if ((num = read(fd, &buf, 6)) == -1) {
			showerr(config->sensors[idx]);
			errcnt++;
			return -1000;
		}
		if (close(fd) < 0) {
			showerr(config->sensors[idx]);
			errcnt++;
			return -1000;
		}
		buf[num] = 0;
		tmp = strtol(buf, &endptr, 10);
		if (tmp > rv) rv = tmp;
		idx++;
	}
	rv = rv/1000;
	if (rv < 1) {
		message(LOG_ERR, MSG_ERR_T_GET);
		errcnt++;
	}
	return rv;
}

/***********************************************************
 * Set fan speed (IBM interface).
 ***********************************************************/
void setfan_ibm(int level) {
	FILE *ibm_fan;
	int ret;

	if ((ibm_fan = fopen(IBM_FAN, "r+")) == NULL) {
		showerr(IBM_FAN);
		errcnt++;
		return;
	}
	ret = fprintf(ibm_fan, "level %d\n", level);
	fclose(ibm_fan);
	if (ret != 8) {
		message(LOG_ERR, MSG_ERR_FAN_WRITE);
		errcnt++;
	}
	return;
}

/***********************************************************
 * Set fan speed (sysfs interface).
 ***********************************************************/
void setfan_sysfs(int level) {
	int fan;
	int ret;
	char *buf = malloc(5 * sizeof(char));

	if ((fan = open(config->fan, O_WRONLY)) < 0) {
		showerr(config->fan);
		errcnt++;
		return;
	}
	ret = snprintf(buf, 5, "%d\n", level);
	ret += write(fan, buf, 5);
	close(fan);
	if (ret < 2) {
		message(LOG_ERR, MSG_ERR_FAN_WRITE);
		errcnt++;
	}
	return;
}

/***********************************************************
 * Suspend/Resume-safe way of setting fan speed
 ***********************************************************/
void setfan_sysfs_safe(int level) {
	init_fan_sysfs();
	setfan_sysfs(level);
}

/*********************************************************
 * Checks for fan_control support in thinkpad_acpi and
 * activates the fan watchdog.
 *********************************************************/
void init_fan_ibm() {
	char *line = NULL;
	size_t count = 0;
	FILE *ibm_fan;
	int module_valid=0;

	if ((ibm_fan = fopen(IBM_FAN, "r+")) == NULL) {
		showerr(IBM_FAN);
		errcnt++;
		return;
	}
	while (getline(&line, &count, ibm_fan) != -1)
		if (!strncmp("commands:", line, 9)) module_valid = 1;
	if (!module_valid) {
		message(LOG_ERR, MSG_ERR_MODOPTS);
		errcnt++;
		return;
	}
	fprintf(ibm_fan, "watchdog %d\n", watchdog_timeout);
	fclose(ibm_fan);
	free(line);
	return;
}

/*********************************************************
 * This just activates userspace PWM control
 *********************************************************/
void init_fan_sysfs() {
	int ret;
	char *fan_enable = (char *) malloc((strlen(config->fan) + 8) * sizeof(char));
	FILE *fan;
	size_t s;

	strcpy(fan_enable, config->fan);
	strcat(fan_enable, "_enable");

	if ((fan = fopen(fan_enable, "r+")) == NULL) {
		showerr(config->fan)
		errcnt++;
		return;
	}
	getline(&oldpwm, &s, fan);
	ret = fprintf(fan, "%d\n", 1);
	fclose(fan);
	free(fan_enable);
	if (ret < 2) {
		showerr(fan_enable)
		errcnt++;
		return;
	}
	return;
}

/*********************************************************
 * Restores automatic fan control.
 *********************************************************/
void uninit_fan_ibm() {
	FILE *fan;

	if ((fan = fopen(IBM_FAN, "r+")) == NULL) {
		showerr(IBM_FAN);
		errcnt++;
	}
	else {
		fprintf(fan, "level auto\n");
		fclose(fan);
	}
}

/*********************************************************
 * Restores previous fan control mode.
 *********************************************************/
void uninit_fan_sysfs() {
	FILE *fan;

	if ((fan = fopen(config->fan, "r+")) == NULL) {
		showerr(config->fan);
		errcnt++;
	}
	else {
		fprintf(fan, "%s", oldpwm);
		fclose(fan);
	}
}


